'' Propeller EEPROM Loader v1.0
'' Copyright  2006 by Michael Green
'' Permission is given to modify or incorporate this into any
'' Propeller program as long as this copyright notice retained.

'' This object functions like the bootstrap loader in the Propeller
'' ROM.  It is based on the information in the object "Propeller
'' Loader v1.0" from Parallax.  It is intended to allow downloads
'' from the Propeller Tool IDE via a USB2SER or similar adapter to
'' any three I/O pins on the Propeller.  Note that the built-in
'' adapter on most Propeller boards cannot be used because the
'' reset line would be triggered by the IDE via the adapter.

'' The received data is copied directly into an EEPROM using the
'' I2C routines that are part of the "Propeller Operating System".

CON
  bufferSize = 64                               ' 64 byte buffer
  timeout = 30                                  ' 30 second timeout
  #9, normalEnd,   connectErr,  i2cioErr,    checksumErr
      shutdownErr, timeoutErr
  #0, Shutdown, LoadRun, ProgramShutdown, ProgramRun

VAR
  byte buffer[bufferSize * 2]

OBJ
  ldr : "OS_loader"
  vga : "OS_vgaHiresText"

PUB start(RESn, rxPin, txPin, address) : cog | time, secs, i, j
  pause5ms     := (clkfreq / 1_000) * 5         ' 5ms pause for EEPROM write
  pause250ms   := (clkfreq / 1_000) * 250       ' 250ms pause for acknowledge
  rxPinNum     := rxPin                         ' Receive pin number for CTRA
  resetPinMask := |< RESn                       ' Reset pin mask
  rxPinMask    := |< rxPin                      ' Receive pin mask
  txPinMask    := |< txPin                      ' Transmit pin mask
  eepromAddr   := address                       ' EEPROM address for writing
  startBuffer  := @buffer                       ' 1st of 2 x 32 byte buffers
  endBuffer1   := startBuffer + bufferSize
  endBuffer2   := endBuffer1 + bufferSize
  version      := chipver                       ' Chip version
  secs := 0                                     ' Keep track of # seconds
  cog := cognew(@download,@cog) + 1             ' Attempt to start cog
  time := cnt
  vga.out(" ")                                  ' ***** ***** ***** Debug
  repeat
    if (cog > 8) or (cog == 0)                  ' Return if stopped
      case cog
        0:
          vga.str(string(" Can't start cog",vga#Cr))
        connectErr:
          vga.str(string(" Connection lost",vga#Cr))
        i2cioErr:
          vga.str(string(" I2C overrun",vga#Cr))
        checksumErr:
          vga.str(string(" Checksum error",vga#Cr))
        shutdownErr:
          vga.str(string(" Told to shutdown",vga#Cr))
        timeoutErr:
          vga.str(string(" Timeout expired",vga#Cr))
        other:
          vga.str(string(" Time = "))
          vga.dec(cog / (clkfreq / 1_000_000))
          vga.str(string("us",vga#Cr))
      return
    if (cnt - time) > clkfreq                   ' Count the number of
      secs++                                    '  seconds (approximately)
      vga.dec(secs//10)                         ' ***** ***** ***** Debug
      time += clkfreq
    if secs => timeout                          ' Check for timeout
      cogstop(cog - 1)                          ' If so, stop the cog and
      dira[txPin]~                              '  change transmit to an input
      return timeoutErr

DAT
'' This routine manages the complete download sequence from detecting
'' calibration pulses to final acknowledgement.  One major assumption is
'' that the time required to do a paged write to EEPROM is less than the
'' time required to read in the next 64 byte page.  This is not checked!
'' Because the last page may contain only one byte, there is an explicit
'' wait for the previous paged write to complete.  The EEPROM address
'' must be a multiple of the EEPROM's paged write page size.  This is at
'' least 64 bytes in 32Kx8 EEPROMs.  The transmit line is left floating
'' on error or normal exit.  Note that there are several places in these
'' routines where a missing pulse can cause an indefinite wait.  This
'' routine should be run with a watchdog timer if this is a concern.
                        org     0
download                mov     temp,#bufferSize        ' Include buffer size in
                        shl     temp,#16                '  various buffer pointers
                        or      startBuffer,temp        '   for coding convenience
                        or      endBuffer1,temp
                        or      endBuffer2,temp
                        call    #debugInit              ' ***** ***** ***** Debug
                        mov     bufferPtr,startBuffer
                        mov     temp,#ldr#ioWriteCmd    ' Include write command
                        shl     temp,#24                '  with EEPROM address
                        or      eepromAddr,temp         '   for coding convenience
                        mov     checkSum,#0
                        andn    DIRA,resetPinMask       ' Reset pin is input
                        andn    DIRA,rxPinMask          ' Receive pin is input
                        or      OUTA,txPinMask          ' Transmit pin is output high
                        or      DIRA,txPinMask
                        mov     lfsr,#"P"               ' Initial shift register
                        mov     oneBitWidth,#0          ' No bit timing known yet
                        waitpne resetPinMask,resetPinMask  ' Wait for DTR low
                        call    #receiveBit             ' Get 1st calibration bit
                        mov     oneBitWidth,rxBitWidth  '  should be short (1)
                        call    #receiveBit             ' Get 2nd calibration bit
                        mov     thresholdWidth,rxBitWidth ' should be long (0)
                        add     thresholdWidth,rxBitWidth
                        add     thresholdWidth,rxBitWidth
                        shr     thresholdWidth,#2       ' Threshold = 3/4 long
                        cmp     oneBitWidth,thresholdWidth wc
                        mov     temp,#connectErr        ' 1st received bit should
                if_nc   jmp     #:errorExit             '  have been a 1 bit
                        mov     byteCount,#250
:readLFSR               test    lfsr,#%00000001    wz   ' Get output of LFSR
                        test    lfsr,#%10110010    wc   ' Parity of feedback bits?
                        rcl     lfsr,#1                 ' Shift LFSR
                        call    #receiveBit             ' Get next received bit
                        mov     temp,#connectErr
         if_nc_and_nz   jmp     #:errorExit             ' If mismatch, we have
           if_c_and_z   jmp     #:errorExit             '  a bad connection
                        djnz    byteCount,#:readLFSR
                        mov     saveTime,CNT            ' ***** ***** ***** Debug
                        mov     byteCount,#250
:writeLFSR              test    lfsr,#%00000001    wz   ' Get output of LFSR
                        test    lfsr,#%10110010    wc   ' Parity of feedback bits?
                        rcl     lfsr,#1                 ' Shift LFSR
                        call    #transmitBit            ' Transmit LFSR bit
                        djnz    byteCount,#:writeLFSR
                        mov     bitCount,#8             ' Transmit an 8 bit chip
                        mov     bitValue,version        '  version number
                        call    #transmitData
                        mov     temp,CNT                ' ***** ***** ***** Debug
                        sub     temp,saveTime           ' ***** ***** ***** Debug
                        jmp     #:errorExit             ' ***** ***** ***** Debug
                        mov     debugChr,#"%"           ' ***** ***** ***** Debug
                        call    #debugOut               ' ***** ***** ***** Debug
                        mov     bitCount,#32            ' Receive a 32 bit command
                        call    #receiveData
                        mov     debugVal,bitValue       ' ***** ***** ***** Debug
                        call    #debugHexLong           ' ***** ***** ***** Debug
                        mov     debugChr,#"?"           ' ***** ***** ***** Debug
                        call    #debugOut               ' ***** ***** ***** Debug
                        mov     command,bitValue
                        mov     temp,#shutdownErr
                        tjz     command,#:errorExit     ' Command == Shutdown
                        mov     bitCount,#32
                        call    #receiveData            ' Receive a 32 bit byte count
                        mov     debugVal,bitValue       ' ***** ***** ***** Debug
                        call    #debugHexLong           ' ***** ***** ***** Debug
                        mov     debugChr,#"#"           ' ***** ***** ***** Debug
                        call    #debugOut               ' ***** ***** ***** Debug
                        mov     byteCount,bitValue
:getNextByte            mov     bitCount,#8
                        call    #receiveData            ' Get the next data byte and
                        wrbyte  bitValue,bufferPtr      '  store it in the buffer
                        add     bufferPtr,#1
                        add     checkSum,bitValue       ' Compute checksum
                        cmp     bufferPtr,endBuffer1 wz ' Finished filling buffer 1?
                if_z    call    #checkCompleteIO
                if_z    wrlong  startBuffer,ioControl1
                if_z    jmp     #:initiateWrite
                        cmp     bufferPtr,endBuffer2 wz ' Finished filling buffer 2?
                if_z    call    #checkCompleteIO
                if_z    wrlong  endBuffer1,ioControl1
                if_z    mov     bufferPtr,startBuffer   ' Start again with buffer 1
:initiateWrite  if_z    wrlong  eepromAddr,ioControl0
                if_z    add     eepromAddr,#32          ' Increment EEPROM address
                if_z    mov     debugChr,#"!"           ' ***** ***** ***** Debug
                if_z    call    #debugOut               ' ***** ***** ***** Debug
                        djnz    byteCount,#:getNextByte
                        test    bufferPtr,#bufferSize-1 wz ' See if any residual bytes
                if_z    jmp     #:acknowledge
                        call    #checkCompleteIO        ' Wait for last I/O to finish
                        mov     temp,pause5ms
                        add     temp,CNT                ' Wait 5ms for write complete
                        waitcnt temp,#0
                        mov     temp,bufferPtr          ' Set up residual byte count
                        and     temp,#bufferSize-1
                        shl     temp,#16
                        cmp     bufferPtr,endBuffer1 wc ' Using 1st or 2nd buffer?
                if_c    or      temp,startBuffer
                if_nc   or      temp,endBuffer1
                        wrlong  temp,ioControl1         ' Initiate last I2C I/O
                        wrlong  eepromAddr,ioControl0                        
                        mov     debugChr,#"!"           ' ***** ***** ***** Debug
                        call    #debugOut               ' ***** ***** ***** Debug
:acknowledge            test    checkSum,#$FF      wz   ' Checksum should be zero
                if_nz   mov     temp,#checksumErr
                if_nz   jmp     #:errorExit
                        mov     debugChr,#"*"           ' ***** ***** ***** Debug
                        call    #debugOut               ' ***** ***** ***** Debug
                        mov     temp,pause250ms
                        add     temp,CNT                ' Wait 250ms
                        waitcnt temp,#0
                        test    temp,#0            wz   ' Transmit acknowledge
                        call    #transmitBit
                        call    #checkCompleteIO        ' Make sure last I/O ok
                        cmp     command,#ProgramShutdown wc
                if_c    jmp     #:normalExit
                        test    temp,#0            wz   ' Transmit acknowledge
                        call    #transmitBit
                        mov     temp,pause250ms
                        add     temp,CNT                ' Wait 250ms
                        waitcnt temp,#0
                        test    temp,#0            wz   ' Transmit acknowledge
                        call    #transmitBit
                        mov     temp,pause250ms
                        add     temp,CNT                ' Wait 250ms
                        waitcnt temp,#0
:normalExit             mov     temp,#normalEnd         ' Mark operation complete
:errorExit              wrlong  temp,PAR
                        andn    DIRA,txPinMask          ' Turn off transmit output
                        cogid   temp                    '  and stop cog
                        cogstop temp

checkCompleteIO         rdlong  temp,ioControl0         ' Make sure previous I/O
                        test    temp,ioMaskRdy     wz   '  hasn't overrun and
                if_nz   mov     temp,#i2cioErr
                if_nz   jmp     #connectErr             '   had no errors
                        test    temp,ioMaskErr     wz
                if_nz   mov     temp,#i2cioErr
                if_nz   jmp     #connectErr             ' Return with (z) true
checkCompleteIO_ret     ret

receiveData             mov     bitMask,#1              ' Receive one or more bits
                        mov     bitValue,#0             '  in LSB..MSB order
:getNextBit             call    #receiveBit
                        muxc    bitValue,bitMask
                        shl     bitMask,#1
                        djnz    bitCount,#:getNextBit
receiveData_ret         ret

receiveBit              waitpne rxPinMask,rxPinMask     ' Wait for start pulse
                        mov     temp,CNT                '  save start time (ticks)
                        andn    OUTA,txPinMask          '   and echo the pulse
                        waitpeq rxPinMask,rxPinMask     ' Wait for end of pulse
                        mov     rxBitWidth,CNT          '  save end time (ticks)
                        or      OUTA,txPinMask          '   and echo the pulse
                        sub     rxBitWidth,temp         ' Compute pulse width
                        cmp     rxBitWidth,thresholdWidth wc ' Is 1 or 0?
receiveBit_ret          ret

transmitData            mov     bitMask,#1              ' Transmit one or more bits
:sendNextBit            test    bitValue,bitMask   wz   '  in LSB..MSB order
                        call    #transmitBit
                        shl     bitMask,#1
                        djnz    bitCount,#:sendNextBit
transmitData_ret        ret

transmitBit             waitpne rxPinMask,rxPinMask     ' Wait for start of pulse
                        mov     temp,CNT                '  and save the start time
                        waitpeq rxPinMask,rxPinMask     ' Wait for end of pulse
                        mov     rxBitWidth,CNT          '  and save the end time
                        sub     rxBitWidth,temp         ' Compute the pulse width
                        cmp     rxBitWidth,thresholdWidth wc
                if_nc   jmp     #transmitBit            ' Ignore long pulses
                        add     rxBitWidth,CNT          ' Now echo the start pulse
                        andn    OUTA,txPinMask          '  (delayed by one bit time)
                        waitcnt rxBitWidth,oneBitWidth
                if_nz   or      OUTA,txPinMask          ' Output the data pulse
                        waitcnt rxBitWidth,oneBitWidth
                        or      OUTA,txPinMask          ' Finish up with stop pulse
                        waitcnt rxBitWidth,#0           '  (past the long pulse end)
transmitBit_ret         ret                


'' Debugging routines that write directly to the display buffer

debugHexPart            mov     debugShift,#0
                        mov     debugChr,#0
:debugHexLoop           rol     debugVal,#4
                        mov     debugDig,debugVal
                        and     debugDig,#%1111
                        add     debugDig,#"0"
                        cmp     debugDig,#"9"+1    wc
              if_nc     add     debugDig,#"A"-("9"+1)
                        shl     debugDig,debugShift
                        or      debugChr,debugDig
                        add     debugShift,#8
                        djnz    debugCnt,#:debugHexLoop
                        call    #debugOut
debugHexPart_ret        ret

debugHexByte            mov     debugCnt,#2
                        shl     debugVal,#24
                        call    #debugHexPart
debugHexByte_ret        ret

debugHexWord            mov     debugCnt,#4
                        shl     debugVal,#16
                        call    #debugHexPart
debugHexWord_ret        ret

debugHexLong            mov     debugHexPart_ret,debugHexLong_ret
                        mov     debugCnt,#4
                        call    #debugHexPart
                        mov     debugCnt,#4
                        call    #debugHexPart
debugHexLong_ret        ret

debugInit               rdword  debugDsp,screenPtrAdr    ' Write directly to the screen
debugInit_ret           ret                              

debugOut                or      debugChr,#0        wz    ' Stop when debugChr all zero
              if_z      jmp     #debugOut_ret
:debugNext              wrbyte  debugChr,debugDsp        ' Write char & move to next
                        add     debugDsp,#1             
                        shr     debugChr,#8        wz    ' Order is LSB..MSB          
              if_nz     jmp     #:debugNext
debugOut_ret            ret

screenPtrAdr            long    ldr#screenPtr            ' Address of screen pointer
debugDsp                long    0                        ' Pointer to display memory
debugChr                long    0                        ' Up to 4 characters to output
debugVal                long    0
debugDig                long    0
debugShift              long    0
debugCnt                long    0
pointer                 long    0-0
endSnap                 long    0-0
saveTime                long    0

wordMask                long    $0000FFFF
addrMask                long    $00FFFFFF
divisorBy3              long    3 << 15
ioMaskRdy               long    ldr#ioTestRdy           ' Masked bits == 0 if ready
ioMaskErr               long    ldr#ioTestErr           ' Masked bit == 0 if no error
cmdCode                 long    ldr#ioWriteCmd << 24    ' Used in I2C control block
ioControl0              long    ldr#ioControl           ' Addresses of 1st and 2nd
ioControl1              long    ldr#ioControl + 4       '  longs in I2C control block
pause5ms                long    0-0                     ' clkfreq / 200
pause250ms              long    0-0                     ' clkfreq / 4
version                 long    0-0                     ' Propeller chip version
txPinMask               long    0-0                     ' Bitmask for transmit pin
rxPinMask               long    0-0                     ' Bitmask for receive pin
rxPinNum                long    0-0                     ' Receive pin number
resetPinMask            long    0-0                     ' Bitmask for reset pin
startBuffer             long    0-0                     ' Address of start of buffers
endBuffer1              long    0-0                     ' Address past end of buffer 1
endBuffer2              long    0-0                     ' Address past end of buffer 2
eepromAddr              long    0-0                     ' EEPROM address & write cmd
oneBitWidth             res     1                       ' # of clock ticks for one bit
thresholdWidth          res     1                       ' Threshold between 0 and 1
rxBitWidth              res     1
checkSum                res     1                       ' Checksum of bytes received
lfsr                    res     1                       ' Linear feedback shift reg
command                 res     1                       ' Command received from IDE
byteCount               res     1                       ' Byte count received from IDE
bufferPtr               res     1
bitCount                res     1
bitValue                res     1
bitMask                 res     1
temp                    res     1
                        fit