'' This object provides declarations and routines to manipulate
'' named object code spaces in I2C EEPROMS.  It also handles a
'' persistent storage area kept in the last long of the boot
'' EEPROM area.  This is overwritten when a new program is
'' downloaded to the EEPROM by the IDE, but is relocated to the
'' next "page" of the boot EEPROM when this is available.  This
'' area keeps a map of which I/O pins are used for I2C busses.
'' If present, the system automatically probes those busses for
'' EEPROMs during initialization.  There is a copy of the bitmap
'' stored in every code space when it is either named or deleted,
'' but the only copy referenced is the one in the second or first
'' code space of the boot EEPROM.

'' Named object code spaces are currently always 32Kx8 with the
'' name of the program stored in the last 4 longs (16 bytes) of
'' the 32Kx8 "page".  The last long contains system information
'' including a checksum so the file name is actually 12 chars.

'' The bitmap of pins with I2C busses is stored in every initialized
'' "page" for checking purposes although only the one in the boot
'' EEPROM or the next 32K page on the boot EEPROM buss is referenced.
'' A uninitialized "page" is one where at least one checksum doesn't
'' match.  A deleted "page" is one where the checksums are valid and
'' the filename is blank.

'' These routines should be able to support another storage scheme,
'' but others may require some form of space compaction which can
'' be slow with a limited speed serial bus.

CON
  busBitmap   = $8000 - 2         ' Bitmap of pins with I2C busses
  checkMap    = $8000 - 3         ' Checksum for busBitmap
  checkSum    = $8000 - 4         ' Checksum for filename
  filename    = $8000 - 16        ' 12 character filename

  bootName    = $544F4F42         ' BOOT
  blankName   = $20202020         ' blanks

  opt_text    = %0000_0000_0001   ' file is text (not binary)

OBJ
  ldr   : "OS_loader"             ' System definitions
  vga   : "OS_vgaHiresText"       ' Hires VGA driver modified for OS use
  
PUB updateBitmapEEPROMs(addI2C) | i, t, m, x
'' This routine scans all the I2C busses identified in the stored
'' bitmap and fills in the EEPROM table based on what's present.
'' Any I2C busses that do not have EEPROMs are removed from the bitmap.
'' The parameter is either -1 or the number of a pin pair to try adding.
  result := getBusBitmap
  if addI2C <> -1
    result |= |<(addI2C >> 1)     ' Add specified bus
  repeat i from 0 to ldr#maxPinPairs-1
    m := |<i
    if (result & m) <> 0          ' If known, check if present now
      t := probeEEPROMs(i<<1)
      if t == 0
        result &= !m              ' If not present, remove from bitmap
        if (i<<1) == ldr#i2cBootSCL
          abort result            ' Boot EEPROM not present
      word[long[ldr#systemPtr]+ldr#sys_eeproms][i] := t
  updateBitmap(result)

PRI getBusBitmap | i, j, mem, chkM, c, bitmap
'' This routine checks the first 32K of the boot EEPROM for a valid
'' I2C pin bitmap.  If there's more than 32K, the second 32K "page"
'' is checked first.  If there's a valid bitmap, this is returned.
'' Otherwise, a default bitmap showing only the boot EEPROM is used.
  mem := word[long[ldr#systemPtr]+ldr#sys_eeproms+ldr#i2cBootSCL]
  repeat i from 1 to 0
    if (i == 0) or ((mem & |<i) <> 0)
      bitmap~
      chkM := 0
      ldr.readEEPROM((ldr#i2cBootSCL<<18)|(i<<15)|checkMap,@chkM,1)
      repeat j from 0 to 1
        c := 0
        ldr.readEEPROM((ldr#i2cBootSCL<<18)|(i<<15)|(busBitmap+j),@c,1)
        bitmap |= c << (8 * j)
        chkM += c
      if (chkM & $FF) == $FF
        return bitmap
  return |<(ldr#i2cBootSCL >> 1)

PRI updateBitmap(bitmap) | i, j, mem, chkM, c
'' This routine updates the I2C bus bitmap stored in the boot
'' EEPROM.  If there's a second "page", the bitmap is recorded
'' there.  Otherwise, it's saved in the first (or only) 32K.
  mem := word[long[ldr#systemPtr]+ldr#sys_eeproms+ldr#i2cBootSCL]
  repeat i from 1 to 0
    if (|<i & mem) or (i == 0)
      chkM := $FF
      repeat j from 0 to 1
        c := bitmap & $FF
        bitmap := bitmap >> 8
        chkM -= c
        ldr.writeEEPROM((ldr#i2cBootSCL<<18)|(i<<15)|(busBitmap+j),@c,1)
        waitcnt(clkfreq / 200 + cnt)
      ldr.writeEEPROM((ldr#i2cBootSCL<<18)|(i<<15)|checkMap,@chkM,1)
      waitcnt(clkfreq / 200 + cnt)
      return

PUB markNewName(namePtr,addr,opt)
'' This routine initializes the code space at the specified address
'' on the I2C bus with the supplied file name.  It does check that
'' the code space is either marked as deleted or is uninitialized.
'' The sign bits of the name are used for 12 option bits given here.
'' The address format is pin pair in bits 22..19, page number in
'' bits 18..15, and address in 14..0.
  addr &= $7FF000                 ' If option bits present, mask them
  if checkSums(0,addr) <> true
    writeFilename(namePtr,addr,opt)
    return false
  abort -string("file not created, space not valid")

PUB markDeleted(namePtr,addr)
'' This routine marks as deleted the code space at the specified
'' address on the I2C bus.  It does check that the code space
'' has the specified name unless the pointer is zero.
'' The sign bits of the name are used for 12 option bits, zeroed here.
'' The address format is pin pair in bits 22..19, page number in
'' bits 18..15, and address in 14..0.
  addr &= $7FF000                 ' If option bits present, mask them
  if checkSums(namePtr,addr) <> true
    writeFilename(0,addr,0)
    return false
  abort -string("file not deleted, space not valid")

PUB searchFile(namePtr) | i, j, b, addr, t, k
'' This routine searches all available EEPROMs for an object code
'' space with the name given.  It returns the pin pair in bits 22-19,
'' the 32K page number in bits 18-15, and the address in bits 14-0.
'' If the code space is not found, an abort is done with the
'' negative address of an error message returned.  If namePtr is
'' zero, this looks for an uninitialized or deleted code space.
'' The sign bits of the name are used for 12 option bits.  These
'' are returned in the least significant 12 bits of the address.
'' The name "BOOT" is a special case for the boot EEPROM.
  if (long[namePtr][0] == bootName) and (long[namePtr][1] == blankName)
    return ldr#i2cBootSCL << 18
  repeat i from 0 to ldr#maxPinPairs-1   ' Pin pairs
    b := word[long[ldr#systemPtr]+ldr#sys_eeproms][i]
    repeat j from 0 to 15        ' 32Kx8 spaces
      if (|<j & b) <> 0          ' Is this one present?
        addr := (i << 19) | (j << 15)  ' Check name,chksums        
        if (t := checkSums(namePtr,addr)) <> true
          return addr | t        ' Match, return location
  abort -string("file not found") ' and option bits

PRI checkSums(namePtr, addr) | i,p,c,eq,chkS,chkM
'' This routine checks the filename and bitmap checksums.
'' If those are correct, it compares the filename to the
'' one supplied (or to all blanks if none supplied) and
'' returns the options bits if all match.  If the checksums
'' fail and no name was supplied, a false is returned (to
'' catch the uninitialized situation).  Otherwise return true.
'' The boot EEPROM will never be found.  It's a special case.
'' In particular, the name "BOOT" is always the boot EEPROM.
  if ((addr >> 18) & $1E) == ldr#i2cBootSCL
    return true                   ' boot EEPROM never found
  if (long[namePtr][0] == bootName) and (long[namePtr][1] == blankName)
    return 0                      ' name BOOT is always boot EEPROM
  result := 0
  addr &= $7FF000                 ' If option bits present, mask them
  chkM := 0
  ldr.readEEPROM(addr+checkMap,@chkM,1)
  repeat i from 0 to 1
    c := 0
    ldr.readEEPROM(addr+busBitmap+i,@c,1)
    chkM += c
  if (p := namePtr) == 0
    p := string("            ")
  eq := true
  chkS := 0
  ldr.readEEPROM(addr+checkSum,@chkS,1)
  repeat i from 0 to 11
    c := 0
    ldr.readEEPROM(addr+filename+i,@c,1)
    chkS += c
    eq := eq and ((c & $7F) == byte[p++])
    result |= (c >> 7) << i
  if ((chkS & $FF) == $FF) and ((chkM & $FF) == $FF)
    if not eq
      result := true
  else
    result := namePtr <> 0

PRI writeFilename(namePtr, addr, opt) | i, p, c, chkS, chkM, m
'' This routine writes the filename and bitmap fields and their checksums
'' to the end of the 32Kx8 space given by "addr".  The option information
'' in the sign bits of the characters is provided in "opt".
'' The name "BOOT" is a special case for the boot EEPROM and is not stored.
  m := getBusBitmap             ' 
  addr &= $7FF000                 ' If option bits present, mask them
  chkM := $FF
  repeat i from 0 to 1            ' Write bitmap bytes & compute checksum
    c := m & $FF
    m := m >> 8
    chkM -= c
    ldr.writeEEPROM(addr+busBitmap+i,@c,1)
    waitcnt(clkfreq / 200 + cnt)
  ldr.writeEEPROM(addr+checkMap,@chkM,1)
  waitcnt(clkfreq / 200 + cnt)    ' Write bitmap checksum
  if (long[namePtr][0] == bootName) and (long[namePtr][1] == blankName)
    return                        ' Only store the bitmap if boot EEPROM
  if (p := namePtr) == 0
    p := string("            ")   ' If no name, use blanks
  chkS := $FF
  repeat i from 0 to 11
    c := (byte[p++] & $7F) | ((opt << 7) & $80)
    opt >>= 1                     ' Insert option bits,
    chkS -= c                     '  compute checksum,
    ldr.writeEEPROM(addr+filename+i,@c,1)
    waitcnt(clkfreq / 200 + cnt)  '   and write name bytes
  ldr.writeEEPROM(addr+checkSum,@chkS,1)
  waitcnt(clkfreq / 200 + cnt)    ' Write name checksum

PUB probeEEPROMs(pin) : bitmap | device, addr, b0, b0a, b1, b1a, val, seedP, t
'' This routine first attempts to see if an EEPROM is present.
'' A series of 32Kx8 address spaces could be provided by one
'' device or several.  This routine builds a bitmap showing which
'' have been provided if any and returns a value suitable for
'' placing in the system's EEPROM table.
  pin &= $1E
  bitmap := 0
  seedP := long[ldr#systemPtr] + ldr#sys_random
  repeat device from 0 to 14 step 2  ' Go through the EEPROM devices
    addr := ?long[seedP] & $7FFF     ' Generate random 32K address
    b0 := 0
    if not ldr.readEEPROM((pin<<18)|(device<<15)|addr,@b0,1) ' Read current data there
      next
    b1 := 0
    if not ldr.readEEPROM((pin<<18)|(device<<15)|$8000|addr,@b1,1) ' in each 32K page
      next
    bitmap |= |<device               ' We have to see what sizes they are
    repeat                           ' Generate a random byte not in either
      val := ?long[seedP] & $FF      '  of the two 32K pages.  They might be
    until (val <> b0) and (val <> b1)  ' the same physical location.
    ldr.writeEEPROM((pin<<18)|(device<<15)|$8000|addr,@val,1) ' Write byte in 2nd 32K page
    waitcnt(clkfreq / 200 + cnt)     ' (5ms for write)
    b0a := 0
    ldr.readEEPROM((pin<<18)|(device<<15)|addr,@b0a,1) ' Now see which 32K page changed
    b1a := 0
    ldr.readEEPROM((pin<<18)|(device<<15)|$8000|addr,@b1a,1)
    if (b0 == b0a) and (b1a == val)  ' If the change is in the second 32K only
      bitmap |= |<(device + 1)       '  then we really have memory there
      ldr.writeEEPROM((pin<<18)|(device<<15)|$8000|addr,@b1,1) ' Restore original data
    else
      ldr.writeEEPROM((pin<<18)|(device<<15)|addr,@b0,1) ' Not there? Restore first 32K
    waitcnt(clkfreq / 200 + cnt)     ' (5ms for write)

PUB dumpMemory(pin,addr,size) | i, c, p, page, first
'' This routine dumps a portion of the RAM (pin > highest pin) or EEPROM to the
'' display.  The format is 16 bytes wide with hexadecimal and printable characters.
'' If no EEPROM is present at that location, an error message is displayed.
  page := (addr >> 15) & $1F
  addr &= $7FFF
  if pin < (ldr#maxPinPairs << 1)
    if not ldr.readEEPROM((pin<<18)|(page<<15)|addr,@c,1)
      vga.str(string("*** no EEPROM at that address",vga#Cr))
      return
  if addr + size > $7FFF               ' truncate to end of code space
    size := $8000 - addr
  first := true
  p := addr & $7FF0
  repeat while p < (addr + size)
    vga.str(string("    "))
    if pin < (ldr#maxPinPairs << 1)
      vga.decFixed(pin,2)
      vga.out(":")
    else
      vga.str(string("   "))
    if first
      vga.hex((page << 15) | addr,5)
      first := false
    else
      vga.hex((page << 15) | p,5)
    vga.out3(vga#MoveX,62,"|")
    repeat i from 0 to 15
      if p => addr and p < (addr + size)
        if pin < (ldr#maxPinPairs << 1)
          ldr.readEEPROM((pin<<18)|(page<<15)|p,@c,1)
        else
          c := byte[p]
        vga.out2(vga#MoveX,3*i+13)
        vga.hex(c,2)
        vga.out2(vga#MoveX,i+63)
        if c < " " or c > "~"
          vga.out(".")
        else
          vga.out(c)
      p++
    vga.out4(vga#MoveX,79,"|",vga#Cr)
  
PUB fileList | i, j, k, bitmap, chkM, chkS, t, opt, a, b, c
'' This routine displays a list of all files currently known to the system.
  repeat i from 0 to 15          ' Pin pairs
    bitmap := word[long[ldr#systemPtr]+ldr#sys_eeproms][i]
    repeat j from 0 to 15        ' 32Kx8 spaces
      if (|<j & bitmap) <> 0     ' Is this one present?
        chkM := 0
        ldr.readEEPROM((i<<19)|(j<<15)|checkMap,@chkM,1)
        repeat k from 0 to 1     ' Make sure bitmap checksum correct
          t := 0
          ldr.readEEPROM((i<<19)|(j<<15)|busBitmap+k,@t,1)
          chkM += t
        chkS := 0
        ldr.readEEPROM((i<<19)|(j<<15)|checkSum,@chkS,1)
        opt := 0                 ' Read filename, extract option bits,
        repeat k from 0 to 11    '  and make sure checksum is correct
          t := 0
          ldr.readEEPROM((i<<19)|(j<<15)|filename+k,@t,1)
          chkS += t
          byte[@a][k] := t & $7F
          opt |= t >> 7 << k
        vga.str(string("    "))
        vga.decFixed(i<<1,2)
        vga.out(":")
        vga.hex(j<<15,5)
        if (chkS & $FF) == $FF and (chkM & $FF) == $FF
          if a == $20202020 and b == $20202020 and c == $20202020
            vga.str(string(" Unused (deleted)",vga#Cr))
          else
            vga.str(string(" <"))
            repeat k from 0 to 11
              vga.out(byte[@a][k])
            vga.str(string("> [$"))
            vga.hex(opt,3)
            if opt & opt_text
              vga.out(",")
              vga.str(string("text"))
            vga.str(string("]",vga#Cr))
        else
          if (i == 14) and (j == 0)
            vga.str(string(" Boot EEPROM",vga#Cr))
          else
            vga.str(string(" Unused (uninitialized)",vga#Cr))