'' This is the initialization (main) routine for the Propeller "Operating System"
'' This set of objects is intended to provide a basic set of I/O including VGA
'' text video, a PS/2 keyboard, and management of one or more I2C memory busses,
'' each containing up to 256K of EEPROM.  One is always the boot I2C bus on I/O
'' pins 28 and 29.  Others can be added on any pair of I/O pins.  The VGA pins
'' and keyboard pins are those used by the Propeller Demo Board, Rev C.  Included
'' is a SPIN loader that can load a normal SPIN program from one of the EEPROMs
'' known to the system, from any location in the EEPROM.  The size of the program
'' is stored with the program by the IDE, but this can be overridden.
'' The I/O drivers are intended to run continuously in 3 COGs (2 for video and 1
'' for the keyboard).  Their work areas (including buffers) are kept at the end
'' of RAM memory so they should not be overwritten when a new program is loaded.
'' The new program, if it uses these video text, keyboard, and loader objects,
'' should be able to seamlessly use the video and keyboard and transfer control
'' further to other programs and back to this main routine for further interaction.
'' The loader will stop any COGs not known to it prior to loading a new progam.
'' Programs can register COGs intended to "persist" between programs.  These
'' must be running assembly language and use a work area that won't be overlaid.
'' There are methods provided by the loader to do simple work area allocation.

'' All routines (except drivers where noted) are (c) 2006 by Michael Green.
'' Some I/O drivers are copyrighted by Parallax, some with modifications by
'' others.  See the notes at the beginning of each source file for specifics.

CON
  _clkmode    = xtal1 + pll16x
  _xinfreq    = 5_000_000

  vgaPinGroup = 16                             ' Base pin for VGA pin group
  ps2KeyData  = 26                             ' Data pin for PS/2 keyboard
  ps2KeyClock = 27                             ' Clock pin for PS/2 keyboard
  
OBJ
  ldr     : "OS_loader"                        ' SPIN program loader and support routines
  ldrInit : "OS_loaderInit"                    '  Initialization and assembly portion
  vga     : "OS_vgaHiresText"                  ' Hires VGA driver modified for OS use
  vgaInit : "OS_vgaHiresTextInit"              '  Initialization and assembly portion
  key     : "OS_keyboard"                      ' PS/2 keyboard driver modified for OS use
  keyInit : "OS_keyboardInit"                  '  Initialization and assembly portion
  files   : "OS_fileSystem"                    ' File system support
  edit    : "OS_windowEditor"                  ' Screen and file editor
  load    : "OS_eepromLoader"                  ' Downloads from Propeller Tool to EEPROM

CON
  _free = (ldr#endMemory - ldr#endFree + ldr#totalNeeded + 3) / 4  ' # longs needed
                                               
  #1, processProbe, processInfo,  processEdit  ' For command post-processing
      processHelp,  processRun,   processList
      processCopy,  processDump,  processDelete
      processTest

  bootAddr   = ldr#i2cBootSCL << 18            ' File address for boot EEPROM

VAR
  long dumpPin, dumpAddr, dumpSize
  long fileAddr, fileAddrTo, fileChange
  long fileName[3], fileNameTo[3]

PUB initialize | t, i, p
  if long[ldr#initMarker] <> ldr#uniqueMark                ' Check for a valid initialization
    repeat i from 0 to 7
      if i <> cogid                                        ' Stop all cogs except this one
        cogstop(i)
    long[ldr#memPtr]    := ldr#endFree                     ' Start at end of HUB memory
    p := ldr.allocatePRI(ldr#systemSize)                   ' Allocate system work area
    long[ldr#systemPtr] := p
    long[p+ldr#sys_cog] := |<cogid                         ' Mark current cog in use
    byte[p+ldr#sys_cogMain] := cogid + 1
    long[p+ldr#sys_random] := ldr#initMarker               ' Initialize random seed
    wordfill(p+ldr#sys_eeproms,0,ldr#maxPinPairs)          ' Clear EEPROMs bitmap table
    long[ldr#videoPtr]  := ldr.allocatePRI(ldr#videoSize)  ' Allocate work area for video
    long[ldr#screenPtr] := ldr.allocatePRI(ldr#screenSize) ' Allocate screen buffer
    long[ldr#colorPtr]  := ldr.allocatePRI(ldr#colorSize)  ' Allocate screen color buffer
    long[ldr#fontPtr]   := ldr.allocatePRI(ldr#fontSize)   ' Allocate font table
    long[ldr#cursorPtr] := ldr.allocatePRI(ldr#cursorSize) ' Allocate cursor control block
    if not vgaInit.start(vgaPinGroup)                      ' Now start up the VGA driver
      abort false
    vga.str(string("Propeller Operating System - Version 1.035",vga#Cr))
    long[ldr#keybdPtr]  := ldr.allocatePRI(ldr#keybdSize)  ' Allocate work area for keyboard
    if not keyInit.start(ps2KeyData,ps2KeyClock)           ' Now start up the keyboard driver
      abort false
    if not ldr.start(ldrInit.getEntryPoint)                ' Now start up the I2C driver
      abort false                                         
    long[ldr#userPtr] := ldr#noSuchAddr
    long[ldr#initMarker] := ldr#uniqueMark                 ' Basic initialization complete
    waitcnt(clkfreq * 2 + cnt)                             ' Wait for keyboard to initialize
    vga.out(vga#Cr)
  if not key.present
    vga.str(string("--> keyboard not present, can't continue"))
    long[ldr#initMarker] := 0
    abort false
  if (\files.updateBitmapEEPROMs(ldr#i2cBootSCL) & $4000) == 0  ' Check boot EEPROM present
    vga.str(string("--> boot eeprom not accessable, can't continue"))
    long[ldr#initMarker] := 0
    abort false
  processCommands

PRI processCommands | t, y, i, j, k, z, b0, b1, b2, b3, b4, b5, b6, b7
'' Go back to the full screen window and finish processing the commands.
'' Note for the "test" commands that there is minimal checking for validity
'' of the command arguments and it is very easy to erase the contents of
'' the "wrong" EEPROM page or change the "wrong" EEPROM location.  That's
'' why these are "test" commands and they will eventually be removed.
  repeat
    y := vga.windowRow
    t := \enterDoCommand(y)
    if t < 0
      vga.out4(vga#UseWin,0,vga#Cursor,0)
      vga.out4(vga#MoveXY,vga#cols,y,vga#Cr)
      vga.str(string("*** "))
      vga.str(-t)
      vga.out(vga#Cr)
    else
      vga.out4(vga#UseWin,0,vga#Cursor,0)
      vga.out4(vga#MoveXY,vga#cols,y,vga#Cr)
      case t
        processProbe:
          if dumpPin < (ldr#maxPinPairs << 1)
            if (\files.updateBitmapEEPROMs(dumpPin) & $4000) == 0
              vga.str(string("*** boot eeprom no longer accessable",vga#Cr))
              return
          repeat i from 0 to (ldr#maxPinPairs<<1)-2 step 2
            z~
            y := word[long[ldr#systemPtr]+ldr#sys_eeproms+i]
            if y <> 0
              repeat j from 0 to 15
                if (|<j & y) <> 0
                  z += 32
              vga.str(string("    Pin "))
              vga.decFixed(i,2)
              vga.str(string("  "))
              vga.decFixed(z,3)
              vga.str(string("K",vga#Cr))
        processInfo:
          vga.str(string("    pbase="))
          ldr.readEEPROM(fileAddr+ldr#spin_pbase,@t,4)
          vga.hex(t,4)
          vga.str(string(", vbase="))
          ldr.readEEPROM(fileAddr+ldr#spin_vbase,@t,4)
          vga.hex(t,4)
          vga.str(string(", dbase="))
          ldr.readEEPROM(fileAddr+ldr#spin_dbase,@t,4)
          vga.hex(t,4)
          vga.str(string(", pcurr="))
          ldr.readEEPROM(fileAddr+ldr#spin_pcurr,@t,4)
          vga.hex(t,4)
          vga.str(string(", dcurr="))
          ldr.readEEPROM(fileAddr+ldr#spin_dcurr,@t,4)
          vga.hex(t,4)
          vga.out(vga#Cr)
        processEdit:
          editWindow
        processHelp:
'' Copy 32K (0-$7FEF) from one file to another.  Create a file if needed.
          vga.str(string("    copy <fileFrom> <fileTo>",vga#Cr))
'' Delete the specified file
          vga.str(string("    delete <file>",vga#Cr))
'' Display the specified HUB RAM locations
          vga.str(string("    dump <addr> <len>",vga#Cr))
'' Display the specified EEPROM locations
          vga.str(string("    dump <pin> <addr> <len>",vga#Cr))
'' Start a simple screen editor.  Quit using the App Key (to the left of Ins)
          vga.str(string("    edit",vga#Cr))
'' Display a list of accepted commands
          vga.str(string("    help",vga#Cr))
'' Show SPIN related pointers for the program in the specified file
          vga.str(string("    info <file>",vga#Cr))
'' List files available including their location and option bits
          vga.str(string("    list",vga#Cr))
'' List all known I2C busses and the total amount of EEPROM space on each
          vga.str(string("    probe",vga#Cr))
'' Probe the possible I2C buss on the specified pin pair for EEPROMs.  List what's known
          vga.str(string("    probe <pin>",vga#Cr))
'' Execute the SPIN program contained in the specified file
          vga.str(string("    run <file>",vga#Cr))
'' Check the checksum of the SPIN program in the specified 32K EEPROM "page"
          vga.str(string("    test checksum <pin> <page>",vga#Cr))
'' Clear (to $FF) the filename area of the specified 32K EEPROM "page"
          vga.str(string("    test clear <pin> <page>",vga#Cr))
'' Fill the specified 32K EEPROM "page" with a known pseudo-random byte sequence
          vga.str(string("    test fill <pin> <page>",vga#Cr))
'' Test that the specified 32K EEPROM "page" is accessible (attempt a read)
          vga.str(string("    test load <pin> <page> <RESn/Rx/Tx>",vga#Cr))
'' Test Propeller Tool to EEPROM downloader
          vga.str(string("    test present <pin> <page>",vga#Cr))
'' Read (and display) the byte from the specified EEPROM location
          vga.str(string("    test read <pin> <page> <address>",vga#Cr))
'' Verify that the 32K EEPROM "page" contains the known pseudo-random byte sequence
          vga.str(string("    test verify <pin> <page>",vga#Cr))
'' Write the specified byte value to the specified EEPROM location
          vga.str(string("    test write <pin> <page> <address> <data>",vga#Cr))
        processList:
          files.fileList
        processDump:
          files.dumpMemory(dumpPin,dumpAddr,dumpSize)
        processCopy:
          if fileChange
            vga.str(string("    Overwrite it? (Y/n) "))
            repeat
              t := key.getkey
              if t == vga#Cr
                t := "y"
            until t == "Y" or t == "y" or t == "N" or t == "n"
            vga.out(t)
            j := 38
          else
            t := "y"
            vga.str(string("    Created"))
            j := 24
          if t == "Y" or t == "y"
            vga.str(string(", Saving ... "))
            repeat i from 0 to $7FE0 step 32
              vga.out2(vga#MoveX,j)
              vga.decFixed(i,5)
              if i == $7FE0
                k := 16
              else
                k := 32
              t := cnt
              repeat until ldr.readEEPROM(fileAddr+i,@b0,k)
                if (cnt - t) > (clkfreq / 200)
                  abort -string("EEPROM read timeout")
              t := cnt
              repeat until ldr.writeEEPROM(fileAddrTo+i,@b0,k)
                if (cnt - t) > (clkfreq / 200)
                  abort -string("EEPROM write timeout")
            vga.out2(vga#MoveX,j)
            vga.str(string("Saved"))
          vga.out(vga#Cr)
        processRun:
          vga.str(string("    Loading ... ",vga#Cr))
          repeat i from 0 to 7  ' Stop any cogs not registered
            if (word[long[ldr#systemPtr]+ldr#sys_cog] & |<i) == 0
              cogstop(i)
          ldr.bootEEPROM(fileAddr)
          vga.str(string("    Error during loading",vga#Cr))
        processDelete:
          vga.str(string("    Are you sure? (Y/n) "))
          repeat
            t := key.getkey
            if t == vga#Cr
              t := "y"
          until t == "Y" or t == "y" or t == "N" or t == "n"
          vga.out(t)
          if t == "Y" or t == "y"
            files.markDeleted(@fileName,fileAddr)
            vga.str(string(" ... Deleted"))
          vga.out(vga#Cr)
        processTest:
          case fileChange
            1:  ' test present <pin> <page>
              vga.str(string("    Address = "))
              vga.hex(fileAddr,6)
              if ldr.checkPresence(fileAddr)
                vga.str(string(" (present)"))
              else
                vga.str(string(" (not present)"))
              vga.out(vga#Cr)
            2:  ' test fill <pin> <page>
              fillAnEEPROM(fileAddr)
            3:  ' test verify <pin> <page>
              checkAnEEPROM(fileAddr)
            4:  ' test read <pin> <page> <addr>
              ldr.readEEPROM(fileAddr,@fileAddrTo,1)
              vga.str(string("    Address = "))
              vga.hex(fileAddr,6)
              vga.str(string(", Data = "))
              vga.hex(fileAddrTo,2)
              vga.out(vga#Cr)
            5:  ' test write <pin> <page> <addr> <data>
              vga.str(string("    Address = "))
              vga.hex(fileAddr,6)
              vga.str(string(", Data = "))
              vga.hex(fileAddrTo,2)
              vga.out(vga#Cr)
              ldr.writeEEPROM(fileAddr,@fileAddrTo,1)
              waitcnt(clkfreq / 200 + cnt) ' 5ms for write
            6:  ' test check <pin> <page>
              checkSumEEPROM(fileAddr)
            7:  ' test clear <pin> <page>
              vga.str(string("    Address = "))
              vga.hex(fileAddr,6)
              vga.out(vga#Cr)
              t := $FF
              repeat i from 0 to 15
                ldr.writeEEPROM(fileAddr+i,@t,1)
                waitcnt(clkfreq / 200 + cnt) ' 5ms for write
            8:  ' test load <pin> <page> <RESn> <Rx> <Tx>
              vga.str(string("    Address = "))
              vga.hex(fileAddr,6)
              vga.str(string(", RESn,Rx,Tx = "))
              vga.dec(dumpPin)
              vga.out(",")
              vga.dec(dumpPin+1)
              vga.out(",")
              vga.dec(dumpPin+2)
              vga.out(vga#Cr)
              load.start(dumpPin,dumpPin+1,dumpPin+2,fileAddr)

PUB editWindow | t
'' Set up a simple screen editor with numbered blank lines.  Use the "App" key
'' to exit the editor (this is left of the "Ins" key on the Parallax keyboard).
  vga.out6(vga#SetWin,1,5,1,vga#cols-5,vga#rows-2)
  vga.out6(vga#SetWin,2,0,1,5,vga#rows-2)
  vga.out3(vga#UseWin,2,vga#ClrScr)
  repeat t from 1 to vga#rows - 2
    vga.out3(vga#MoveXY,0,t-1)
    vga.decFixed(t,4)
  vga.out5(vga#UseWin,1,vga#Cursor,%01111,vga#ClrScr)
  repeat
    t := edit.windowEditor
  until t == $CC
  vga.out5(vga#UseWin,0,vga#Cursor,%00000,vga#ClrScr)

PUB enterDoCommand(y) | t, x, a, b, c
'' Set up a window at the line specified and accept a command
'' line terminated by a Cr.  Scan for a command word and
'' command specific parameters, either words or numbers.
'' If the command is known and the parameters are valid,
'' execute the command.  Abort with a negative error string
'' address or return a positive command specific value.
  vga.out6(vga#SetWin,1,1,y,vga#cols-1,1)
  vga.out6(vga#UseWin,0,vga#MoveXY,0,y,">")
  vga.out5(vga#UseWin,1,vga#ClrScr,vga#Cursor,%01111)
  repeat
    t := edit.windowEditor
  until t == vga#Cr
  x := \getTextWord(0,@a,4,true," ")
  if x < 0
    abort -string("no command given")
  case a
    $59504F43:  ' copy
      x := \getTextWord(x,@fileName,12,true," ")
      if x < 0
        abort -string("missing <fileFrom>")
      x := \getTextWord(x,@fileNameTo,12,true," ")
      if x < 0
        abort -string("missing <fileTo>")
      fileAddr := \files.searchFile(@fileName)
      if fileAddr < 0
        abort -string("<fileFrom> not found")
      fileChange := true
      fileAddrTo := \files.searchFile(@fileNameTo)
      if fileAddrTo < 0
        fileAddrTo := \files.searchFile(0)
        if fileAddrTo < 0
          abort -string("can't create <fileTo>")
        files.markNewName(@fileNameTo,fileAddrTo,0)
        fileChange := false
      return processCopy
    $424F5250:  ' probe
      x := \getNumber(x,@dumpPin)
      if x < 0
        dumpPin := ldr#maxPinPairs << 1
        return processProbe
      if (dumpPin < (ldr#maxPinPairs << 1)) and (dumpPin & 1 == 0)
        return processProbe
      else
        abort -string("pin number odd or does not exist")
    $204E5552:  ' run
      if (x := \getTextWord(x,@fileName,12,true," ")) < 0
        abort -string("missing <file>")
      fileAddr := \files.searchFile(@fileName)
      if fileAddr < 0
        abort -string("<file> not found")
      return processRun
    $54494445:  ' edit
      return processEdit
    $4F464E49:  ' info
      x := \getTextWord(x,@fileName,12,true," ")
      if x < 0
        abort -string("missing <file>")
      fileAddr := \files.searchFile(@fileName)
      if fileAddr < 0
        abort -string("<file> not found")
      return processInfo
    $504C4548:  ' help
      return processHelp
    $5453494C:  ' list
      return processList
    $454C4544:  ' delete
      x := \getTextWord(x,@fileName,12,true," ")
      if x < 0
        abort -string("missing <file>")
      fileAddr := \files.searchFile(@fileName)
      if fileAddr < 0
        abort -string("<file> not found")
      if fileAddr == (ldr#i2cBootSCL << 18)
        abort -string("can't delete boot EEPROM")
      return processDelete
    $504D5544:  ' dump
      x := \getNumber(x,@a)
      if x < 0
        abort -string("missing <pin> or <addr>")
      x := \getNumber(x,@b)
      if x < 0
        abort -string("missing <addr> or <len>")
      x := \getNumber(x,@c)
      if x < 0
        if (a > $7FFFF) or (b => 1024)
          abort -string("invalid number")
        dumpPin := ldr#maxPinPairs << 1
        dumpAddr := a
        dumpSize := b
        return processDump
      else
        if (a => (ldr#maxPinPairs << 1)) or (a & 1) or (b > $7FFFF) or (c => 1024)
          abort -string("invalid number")
        dumpPin := a
        dumpAddr := b
        dumpSize := c
        return processDump
    $54534554:  ' test
      x := \getTextWord(x,@a,4,true," ")
      if x < 0
        abort -string("missing <option>")
      x := \getNumber(x,@b)
      if x < 0
        abort -string("missing <pin>")
      x := \getNumber(x,@c)
      if x < 0
        abort -string("missing <page>")
      if (b => (ldr#maxPinPairs << 1)) or (b & 1) or (c => 16)
        abort -string("invalid number")
      case a
        $53455250:  ' present
          a := 0
          fileChange := 1
        $4C4C4946:  ' fill
          a := 0
          fileChange := 2
        $49524556:  ' verify
          a := 0
          fileChange := 3
        $44414552:  ' read
          x := \getNumber(x,@a)
          if x < 0
            abort -string("missing <addr>")
          a &= $FFFF
          fileChange := 4
        $54495257:  ' write
          x := \getNumber(x,@a)
          if x < 0
            abort -string("missing <addr>")
          a &= $FFFF
          x := \getNumber(x,@fileAddrTo)
          if x < 0
            abort -string("missing <value>")
          fileAddrTo &= $FF
          fileChange := 5
        $43454843:  ' check
          a := 0
          fileChange := 6
        $41454C43:  ' clear
          a := $7FF0
          fileChange := 7
        $44414F4C:  ' load
          x := \getNumber(x,@dumpPin)
          if x < 0
            abort -string("missing <RESn/Rx/Tx>")
          a := 0
          fileChange := 8
        other:
          abort -string("unknown <option>")
      fileAddr := (b << 18) | (c << 15) | a
      return processTest
    other:
      abort -string("unknown command")

PRI skipBlanks(x)
'' Skip blanks until at right border of window and return x
'' of first non-blank or abort with a negative error string
'' address.
  repeat
    if vga.windowAtEnd(x,0)
      abort -string("looking for non-blank at end of line")
  until vga.windowChar(x++,0) <> " "
  return x - 1

PRI getTextWord(x,textAddr,maxSize,uppercase,clearFirst) | c
'' Skip leading blanks, then copy up to maxSize non-blank
'' characters into the area at textAddr.  Return the
'' position of the next blank.  If uppercase is true,
'' store any lower case alphabetic characters as upper
'' case.  If clearFirst is not -1 (true), clear the
'' textAddr area to the value given.  If an error occurs,
'' abort with a negative error string address
  if clearFirst <> -1
    repeat c from 1 to maxSize
      byte[textAddr + c - 1] := clearFirst
  x := skipBlanks(x)
  repeat while (c := vga.windowChar(x++,0)) <> " "
    if maxSize > 0
      if uppercase and c => "a" and c =< "z"
        c := c & !%0100000
      byte[textAddr++] := c
      maxSize--
  return x - 1

PRI getNumber(x,numberAddr) | negative, c, t
'' Skip leading blanks, then convert a decimal number
'' with an optional sign ("+"/"-") into binary to be
'' stored in the long at numberAddr.  Return the
'' position of the next blank.  If an error occurs,
'' abort with a negative error string address.
'' Currently, there's no checking for 32 bit overflow.
  x := skipBlanks(x)
  c := vga.windowChar(x++,0)
  negative := c == "-"
  if negative or c == "+"
    c := vga.windowChar(x++,0)
  if c == "$"
    c := vga.windowChar(x++,0)
    t := hexDigit(c)
    if t < 0
      abort -string("invalid hex character")
    long[numberAddr] := t
    c := vga.windowChar(x++,0)
    repeat until (t := hexDigit(c)) < 0
      long[numberAddr] := long[numberAddr] << 4 | t
      c := vga.windowChar(x++,0)
    if negative
      long[numberAddr] := -long[numberAddr]
    return x - 1
  if c == "%"
    c := vga.windowChar(x++,0)
    if not (c == "0" or c == "1")
      abort -string("invalid hex character")
    long[numberAddr] := c - "0"
    c := vga.windowChar(x++,0)
    repeat while c == "0" or c == "1"
      long[numberAddr] := long[numberAddr] << 1 | (c - "0")
      c := vga.windowChar(x++,0)
    if negative
      long[numberAddr] := -long[numberAddr]
    return x - 1
  if not (c => "0" and c=< "9")
    abort -string("invalid numeric character")
  long[numberAddr] := c - "0"
  c := vga.windowChar(x++,0)
  repeat while c => "0" and c =< "9"
    long[numberAddr] := long[numberAddr] * 10 + c - "0"
    c := vga.windowChar(x++,0)
  if negative
    long[numberAddr] := -long[numberAddr]
  return x - 1

PRI hexDigit(c)
'' Convert a hexadecimal character to the corresponding value or -1 if invalid.
  if c => "0" and c =< "9"
    return c - "0"
  if c => "A" and c =< "F"
    return c - "A" + 10
  if c => "a" and c =< "f"
    return c - "a" + 10
  return -1

PRI fillAnEEPROM(addr) | i, j, k, seed, t, b0, b1, b2, b3, b4, b5, b6, b7
'' Fill a 32K EEPROM page with pseudo-random data with a known initial seed
'' so that the data can be checked when read.  Modified to use block writes
  seed := ldr#uniqueMark
  vga.str(string("    Filling ... "))
  repeat i from 0 to $7FE0 step 32
    vga.out2(vga#MoveX,16)
    vga.decFixed(i,5)
    repeat j from 0 to 31
      byte[@b0][j] := ?seed
    if i == $7FE0             ' Last 32 byte page is only 1/2 filled
      k := 16
    else
      k := 32
    t := cnt
    repeat until ldr.checkPresence(addr+i)     ' Wait for previous writes 
      if (cnt - t) > (clkfreq / 200)
        abort -string("EEPROM write timeout")
    ldr.writeEEPROM(addr+i,@b0,k)
  vga.str(string(vga#MoveX,16,"Done",vga#Cr))

PRI checkAnEEPROM(addr) | i, j, k, v, seed, b0, b1, b2, b3, b4, b5, b6, b7
'' Check a 32K EEPROM page with pseudo-random data from an known initial seed.
'' Report the first error during checking.  Modified to use block reads.
  seed := ldr#uniqueMark
  vga.str(string("    Checking ... "))
  repeat i from 0 to $7FE0 step 32
    vga.out2(vga#MoveX,17)
    vga.decFixed(i,5)
    if i == $7FE0             ' Last 32 byte page is only 1/2 filled
      k := 16
    else
      k := 32
    ldr.readEEPROM(addr+i,@b0,k)
    repeat j from 0 to k-1
      if byte[@b0][j] <> (v := ?seed & $FF)
        vga.str(string(" @ $"))
        vga.hex(i+j,4)
        vga.str(string(" $"))
        vga.hex(v,2)
        vga.str(string(" <> $"))
        vga.hex(byte[@b0][j],2)
        vga.out(vga#Cr)
        return
  vga.str(string(" Done",vga#Cr))

PRI checkSumEEPROM(addr) | i, j, k, v, b0, b1, b2, b3, b4, b5, b6, b7
'' Compute the SPIN checksum for the specified 32K EEPROM "page" and
'' display whether it is valid.  The filename area is not included.
  vga.str(string("    Checksum ... "))
  repeat i from 0 to $7FE0 step 32
    vga.out2(vga#MoveX,17)
    vga.decFixed(i,5)
    v := 0
    if i == $7FE0             ' Last 32 byte page is only 1/2 filled
      k := 16
    else
      k := 32
    ldr.readEEPROM(addr+i,@b0,k)
    repeat j from 0 to k-1
      v += byte[@b0][j]
  if v <> 0
    vga.str(string(" Fail "))
    vga.hex(v,2)
    vga.out(vga#Cr)
    return
  vga.str(string(" Done",vga#Cr))
 